/** Copyright (C) 2017 by Bundesanstalt für Gewässerkunde
 * Software engineering by
 *  Björnsen Beratende Ingenieure GmbH
 *  Dr. Schumacher Ingenieurbüro für Wasser und Umwelt
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.artifacts.sinfo.common;

import java.util.HashSet;
import java.util.Set;

import org.dive4elements.artifactdatabase.state.Facet;
import org.dive4elements.artifacts.CallContext;
import org.dive4elements.river.artifacts.common.AbstractCalculationResult;
import org.dive4elements.river.artifacts.resources.Resources;
import org.dive4elements.river.artifacts.sinfo.flood_duration.FloodDurationCalculationResult;
import org.dive4elements.river.artifacts.states.DefaultState.ComputeType;
import org.dive4elements.river.exports.LongitudinalSectionGenerator;

/**
 * Processor to generate the facet and data series of infrastructure flood heights
 *
 * @author Matthias Schäfer
 *
 */
public final class FloodHeightProcessor extends AbstractSInfoLineProcessor<AbstractCalculationResult> {

    public static final String FACET_FLOOD_HEIGHT = "sinfo_facet_flood_height";

    public static final String FACET_FLOOD_HEIGHT_DESCRIPTION = "sinfo_facet_flood_height.description";

    public static final String FACET_MAIN_VALUE_1_HEIGHT = "mainvalue.1.w";

    public static final String FACET_MAIN_VALUE_2_HEIGHT = "mainvalue.2.w";

    public static final String FACET_MAIN_VALUE_3_HEIGHT = "mainvalue.3.w";

    public static final String FACET_MAIN_VALUE_HEIGHT_DESCRIPTION = "mainvalue.w.description";

    private static final String I18N_AXIS_LABEL = LongitudinalSectionGenerator.I18N_YAXIS_LABEL;

    private static final Set<String> HANDLED_FACET_TYPES = new HashSet<>();

    static {
        HANDLED_FACET_TYPES.add(FACET_FLOOD_HEIGHT);
        HANDLED_FACET_TYPES.add(FACET_MAIN_VALUE_1_HEIGHT);
        HANDLED_FACET_TYPES.add(FACET_MAIN_VALUE_2_HEIGHT);
        HANDLED_FACET_TYPES.add(FACET_MAIN_VALUE_3_HEIGHT);
    }

    public FloodHeightProcessor() {
        super(I18N_AXIS_LABEL, HANDLED_FACET_TYPES);
    }


    @Override
    protected double[][] doGetPoints(final AbstractCalculationResult data, final String facetName) {
        if (FACET_FLOOD_HEIGHT.contentEquals(facetName))
            return ((FloodDurationCalculationResult) data).fetchInfrastructurePoints(SInfoResultType.infrastructureHeight);

        if (FACET_MAIN_VALUE_1_HEIGHT.contentEquals(facetName))
            return ((FloodDurationCalculationResult) data).fetchMainValuePoints(SInfoResultType.waterlevel1);

        if (FACET_MAIN_VALUE_2_HEIGHT.contentEquals(facetName))
            return ((FloodDurationCalculationResult) data).fetchMainValuePoints(SInfoResultType.waterlevel2);

        if (FACET_MAIN_VALUE_3_HEIGHT.contentEquals(facetName))
            return ((FloodDurationCalculationResult) data).fetchMainValuePoints(SInfoResultType.waterlevel3);

        final String error = String.format("Unknown facet name: %s", facetName);
        throw new UnsupportedOperationException(error);
    }

    public static Facet createFloodHeightFacet(final CallContext context, final String hash, final String id,
            final AbstractCalculationResult result, final int index) {
        return AbstractSInfoLineProcessor.createFacet(context, hash, id, result, index, I18N_AXIS_LABEL,
                FACET_FLOOD_HEIGHT, FACET_FLOOD_HEIGHT_DESCRIPTION);
    }

    public static Facet createMainValueHeightFacet(final CallContext context, final String hash, final String id,
            final FloodDurationCalculationResult result, final int index) {
        final String description = Resources.getMsg(context.getMeta(), FACET_MAIN_VALUE_HEIGHT_DESCRIPTION, FACET_MAIN_VALUE_HEIGHT_DESCRIPTION,
                result.getMainValueLabel(index));
        assert ((index >= 0) && (index <= 2));
        final String facetName = new String[] { FACET_MAIN_VALUE_1_HEIGHT, FACET_MAIN_VALUE_2_HEIGHT, FACET_MAIN_VALUE_3_HEIGHT }[index];
        return new SInfoResultFacet(0, facetName, description, I18N_AXIS_LABEL, ComputeType.ADVANCE, id, hash);
    }
}